﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.Reflection;
using System.Windows.Data;

namespace PropertyViewerDemonstration
{
    /// <summary>
    /// IValueConverter implementation that expands some/all of an object's
    /// public properties as a collection of bindable name/value instances.
    /// </summary>
    /// <remarks>
    /// * Bindable instances are of type PropertyDetails, a contained class
    ///   that exposes a property's Name and Value as properties.
    /// * All public properties are enumerated by default; ConverterParameter
    ///   can be used to specify which properties will be enumerated (and in
    ///   which order) by passing a space-delimited list of names.
    /// * If DisplayNameAttribute is associated with a property, DisplayName
    ///   will be used instead of the property name.
    /// * PropertyViewer can be used for simple data visualization, debugging
    ///   of Bindings, and more.
    /// </remarks>
    /// <example>
    /// For ItemsControl/ListBox (accessing the object via DataContext):
    ///   ItemsSource="{Binding Converter={StaticResource PropertyViewer}}"
    /// As above, but with a custom property list:
    ///   ItemsSource="{Binding Converter={StaticResource PropertyViewer},
    ///     ConverterParameter='PropertA PropertyB'}}"
    /// For ItemsControl/ListBox, but specifying the object via Source:
    ///   ItemsSource="{Binding Source={StaticResource ObjectInstance}
    ///     Converter={StaticResource PropertyViewer}}"
    /// </example>
    public class PropertyViewer : IValueConverter
    {
        /// <summary>
        /// Implements IValueConverter.Convert.
        /// </summary>
        public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
        {
            // Fetch parameters
            var propertyInfos = (null != value) ? value.GetType().GetProperties() : new PropertyInfo[0];
            var propertyNames = (null != parameter) ? parameter.ToString().Split() : new string[0];
            // Prepare a collection to return
            var propertyDetails = new List<PropertyDetails>();
            // No names specified (or no value/properties available)
            if ((0 == propertyNames.Length) || (0 == propertyInfos.Length))
            {
                // Add all properties to the collection
                foreach (var propertyInfo in propertyInfos)
                {
                    propertyDetails.Add(new PropertyDetails(GetPropertyName(propertyInfo), GetPropertyValue(propertyInfo, value)));
                }
            }
            else
            {
                // For each property name specified
                foreach (var propertyName in propertyNames)
                {
                    // Try to match the name against each property in turn
                    bool found = false;
                    foreach (var propertyInfo in propertyInfos)
                    {
                        if (propertyName == propertyInfo.Name)
                        {
                            // Match; add the property to the collection
                            propertyDetails.Add(new PropertyDetails(GetPropertyName(propertyInfo), GetPropertyValue(propertyInfo, value)));
                            found = true;
                            break;
                        }
                    }
                    if (!found && (0 < propertyName.Length))
                    {
                        // No matches (and not an empty name); throw an exception
                        throw new ArgumentException("Property \"" + propertyName + "\" does not exist on object " + value + ".");
                    }
                }
            }
            // Return the collection
            return propertyDetails;
        }

        /// <summary>
        /// Implements IValueConverter.ConvertBack.
        /// </summary>
        public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException("PropertyViewer does not use ConvertBack.");
        }

        /// <summary>
        /// Gets the name of the specified property.
        /// </summary>
        /// <param name="propertyInfo">property</param>
        /// <returns>name</returns>
        private static string GetPropertyName(PropertyInfo propertyInfo)
        {
            // Look for DisplayNameAttribute
            var displayNameAttributes = propertyInfo.GetCustomAttributes(typeof(DisplayNameAttribute), true);
            // Return first DisplayNameAttribute.DisplayName or property name if none present
            return (0 < displayNameAttributes.Length) ? ((DisplayNameAttribute)(displayNameAttributes[0])).DisplayName : propertyInfo.Name;
        }

        /// <summary>
        /// Gets the value of the specified property for an instance.
        /// </summary>
        /// <param name="propertyInfo">property</param>
        /// <param name="instance">instance</param>
        /// <returns>value</returns>
        private static object GetPropertyValue(PropertyInfo propertyInfo, object instance)
        {
            try
            {
                // Return instance's value
                return propertyInfo.GetValue(instance, null);
            }
            catch (TargetParameterCountException)
            {
                // Likely an indexer (ex: Array.Items); skip complex properties here
                return null;
            }
        }

        /// <summary>
        /// Bindable object returned by PropertyViewer to represent a property.
        /// </summary>
#if SILVERLIGHT  // Public visibility so Silverlight can see and bind to the class
        public
#endif
        class PropertyDetails
        {
            /// <summary>
            /// Property name.
            /// </summary>
            public string Name { get; private set; }

            /// <summary>
            /// Instance property value.
            /// </summary>
            public object Value { get; private set; }

            /// <summary>
            /// Constructs a PropertyDetails instance.
            /// </summary>
            /// <param name="name">property name</param>
            /// <param name="value">instance property value</param>
            public PropertyDetails(string name, object value)
            {
                Name = name;
                Value = value;
            }

            /// <summary>
            /// Implements Object.ToString to return a meaningful representation.
            /// </summary>
            /// <returns>name: value</returns>
            public override string ToString()
            {
                return Name + ": " + Value;
            }
        }
    }
}

#if SILVERLIGHT
namespace System.ComponentModel
{
    /// <summary>
    /// Simple implementation of DisplayNameAttribute for Silverlight where it
    /// doesn't exist as part of the Framework (as of Version 2 Beta 1).
    /// </summary>
    public class DisplayNameAttribute : Attribute
    {
        public string DisplayName { get; private set; }
        public DisplayNameAttribute(string displayName)
        {
            DisplayName = displayName;
        }
    }
}
#endif
